/*
 * Name:    d_cryostream600_status.c
 *
 * Purpose: MX driver for the Oxford Cryosystems 600 series Cryostream
 *          controller.
 *
 *          Please note that this driver only reads status values.  It does
 *          not attempt to change the temperature settings.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_rs232.h"
#include "mx_analog_input.h"
#include "d_cryostream600_status.h"

MX_RECORD_FUNCTION_LIST mxd_cryostream600_status_record_function_list = {
	NULL,
	mxd_cryostream600_status_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

MX_ANALOG_INPUT_FUNCTION_LIST
	mxd_cryostream600_status_analog_input_function_list =
{
	mxd_cryostream600_status_read
};

MX_RECORD_FIELD_DEFAULTS mxd_cryostream600_status_field_default[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_DOUBLE_ANALOG_INPUT_STANDARD_FIELDS,
	MX_ANALOG_INPUT_STANDARD_FIELDS,
	MXD_CRYOSTREAM600_STATUS_STANDARD_FIELDS
};

long mxd_cryostream600_status_num_record_fields
		= sizeof( mxd_cryostream600_status_field_default )
		/ sizeof( mxd_cryostream600_status_field_default[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_cryostream600_status_rfield_def_ptr
			= &mxd_cryostream600_status_field_default[0];

#define CRYOSTREAM600_DEBUG	FALSE

/* ===== */

MX_EXPORT mx_status_type
mxd_cryostream600_status_create_record_structures( MX_RECORD *record )
{
        const char fname[] =
		"mxd_cryostream600_status_create_record_structures()";

        MX_ANALOG_INPUT *analog_input;
        MX_CRYOSTREAM600_STATUS *cryostream600_status;

        /* Allocate memory for the necessary structures. */

        analog_input = (MX_ANALOG_INPUT *) malloc(sizeof(MX_ANALOG_INPUT));

        if ( analog_input == (MX_ANALOG_INPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_INPUT structure." );
        }

        cryostream600_status = (MX_CRYOSTREAM600_STATUS *)
				malloc( sizeof(MX_CRYOSTREAM600_STATUS) );

        if ( cryostream600_status == (MX_CRYOSTREAM600_STATUS *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
               "Can't allocate memory for MX_CRYOSTREAM600_STATUS structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_input;
        record->record_type_struct = cryostream600_status;
        record->class_specific_function_list
			= &mxd_cryostream600_status_analog_input_function_list;

        analog_input->record = record;

	/* Raw analog input values are stored as doubles. */

	analog_input->subclass = MXT_AIN_DOUBLE;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_cryostream600_status_read( MX_ANALOG_INPUT *ainput )
{
	const char fname[] = "mxd_cryostream600_status_read()";

	MX_CRYOSTREAM600_STATUS *cryostream600_status;
	char command[80];
	char response[100];
	char identifier[10], mode[10], terminator[10];
	double set_temperature, temperature_error, final_temperature;
	double ramp_rate, evaporator_temperature;
	int num_items, ice_block;
	mx_status_type mx_status;

	if ( ainput == (MX_ANALOG_INPUT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_ANALOG_INPUT pointer passed was NULL." );
	}
	if ( ainput->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_ANALOG_INPUT pointer passed was NULL." );
	}

	cryostream600_status =
		(MX_CRYOSTREAM600_STATUS *) ainput->record->record_type_struct;

	if ( cryostream600_status == (MX_CRYOSTREAM600_STATUS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_CRYOSTREAM600_STATUS pointer for analog input '%s' is NULL",
			ainput->record->name );
	}

	/* Request the status of the controller. */

	strcpy( command, "S$" );

	MX_DEBUG(-2,("%s: sending command '%s'.", fname, command));

	mx_status = mx_rs232_putline( cryostream600_status->rs232_record,
					command, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_rs232_getline( cryostream600_status->rs232_record,
					response, sizeof( response ),
					NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s: received response '%s'.", fname, response));

	num_items = sscanf( response, "%s:%s:%lg:%lg:%lg:%lg:%lg:B%d:%s",
			identifier, mode, &set_temperature, &temperature_error,
			&final_temperature, &ramp_rate, &evaporator_temperature,
			&ice_block, terminator );

	if ( num_items != 9 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The response to the command '%s' was not in a "
		"recognizable format.  Response = '%s'",
			command, response );
	}

	MX_DEBUG(-2,("%s: identifier = '%s'", fname, identifier));
	MX_DEBUG(-2,("%s: mode = '%s'", fname, mode));
	MX_DEBUG(-2,("%s: set_temperature = %g", fname, set_temperature));
	MX_DEBUG(-2,("%s: temperature_error = %g", fname, temperature_error));
	MX_DEBUG(-2,("%s: final_temperature = %g", fname, final_temperature));
	MX_DEBUG(-2,("%s: ramp_rate = %g", fname, ramp_rate));
	MX_DEBUG(-2,("%s: evaporator_temperature = %g",
					fname, evaporator_temperature));
	MX_DEBUG(-2,("%s: ice_block = %d", fname, ice_block));
	MX_DEBUG(-2,("%s: terminator = '%s'", fname, terminator));

	switch( cryostream600_status->parameter_type ) {
	case MXT_CRYOSTREAM600_TEMPERATURE:
		ainput->raw_value.double_value =
				set_temperature + temperature_error;
		break;
	case MXT_CRYOSTREAM600_SET_TEMPERATURE:
		ainput->raw_value.double_value = set_temperature;
		break;
	case MXT_CRYOSTREAM600_TEMPERATURE_ERROR:
		ainput->raw_value.double_value = temperature_error;
		break;
	case MXT_CRYOSTREAM600_FINAL_TEMPERATURE:
		ainput->raw_value.double_value = final_temperature;
		break;
	case MXT_CRYOSTREAM600_RAMP_RATE:
		ainput->raw_value.double_value = ramp_rate;
		break;
	case MXT_CRYOSTREAM600_EVAPORATOR_TEMPERATURE:
		ainput->raw_value.double_value = evaporator_temperature;
		break;
	case MXT_CRYOSTREAM600_ICE_BLOCK:
		ainput->raw_value.double_value = (double) ice_block;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Unrecognized parameter type %d for 'cryostream600_status' record '%s'.  "
"The allowed values are in the range (1-7).",
			cryostream600_status->parameter_type,
			ainput->record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

