/*
 * Name:    d_sis3807.c
 *
 * Purpose: MX pulse generator driver for a single output channel of
 *          the Struck SIS3807.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_driver.h"
#include "mx_clock.h"
#include "mx_vme.h"
#include "mx_pulse_generator.h"

#include "i_sis3807.h"
#include "d_sis3807.h"

MX_RECORD_FUNCTION_LIST mxd_sis3807_record_function_list = {
	NULL,
	mxd_sis3807_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxd_sis3807_open,
	NULL,
	NULL,
	NULL
};

MX_PULSE_GENERATOR_FUNCTION_LIST mxd_sis3807_pulse_generator_function_list = {
	mxd_sis3807_busy,
	mxd_sis3807_start,
	mxd_sis3807_stop,
	mxd_sis3807_get_parameter,
	mxd_sis3807_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_sis3807_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_PULSE_GENERATOR_STANDARD_FIELDS,
	MXD_SIS3807_STANDARD_FIELDS
};

long mxd_sis3807_num_record_fields
		= sizeof( mxd_sis3807_record_field_defaults )
			/ sizeof( mxd_sis3807_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_sis3807_rfield_def_ptr
			= &mxd_sis3807_record_field_defaults[0];

/* --- */

static mx_status_type
mxd_sis3807_get_pointers( MX_PULSE_GENERATOR *pulse_generator,
			MX_SIS3807_PULSER **sis3807_pulser,
			MX_SIS3807 **sis3807,
			const char *calling_fname )
{
	const char fname[] = "mxd_sis3807_get_pointers()";

	MX_RECORD *sis3807_pulser_record, *sis3807_record;
	MX_SIS3807_PULSER *local_sis3807_pulser;

	if ( pulse_generator == (MX_PULSE_GENERATOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PULSE_GENERATOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	sis3807_pulser_record = pulse_generator->record;

	if ( sis3807_pulser_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for the MX_PULSE_GENERATOR pointer "
		"passed by '%s' is NULL.", calling_fname );
	}

	local_sis3807_pulser = (MX_SIS3807_PULSER *)
				sis3807_pulser_record->record_type_struct;

	if ( sis3807_pulser != (MX_SIS3807_PULSER **) NULL ) {
		*sis3807_pulser = local_sis3807_pulser;
	}

	if ( sis3807 != (MX_SIS3807 **) NULL ) {
		sis3807_record = local_sis3807_pulser->sis3807_record;

		if ( sis3807_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The sis3807_record pointer for 'sis3807_pulser' "
			"record '%s' passed by '%s' is NULL.",
				sis3807_pulser_record->name,
				calling_fname );
		}

		*sis3807 = (MX_SIS3807 *) sis3807_record->record_type_struct;

		if ( *sis3807 == (MX_SIS3807 *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_SIS3807 pointer for sis3807 record '%s' "
			"used by 'sis3807_pulser' record '%s' "
			"passed by '%s' is NULL.",
				sis3807_record->name,
				sis3807_pulser_record->name,
				calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxd_sis3807_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_sis3807_create_record_structures()";

	MX_PULSE_GENERATOR *pulse_generator;
	MX_SIS3807_PULSER *sis3807_pulser;

	pulse_generator = (MX_PULSE_GENERATOR *)
				malloc( sizeof(MX_PULSE_GENERATOR) );

	if ( pulse_generator == (MX_PULSE_GENERATOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate memory for a MX_PULSE_GENERATOR structure "
		"for record '%s'.", record->name );
	}

	sis3807_pulser = (MX_SIS3807_PULSER *)
				malloc( sizeof(MX_SIS3807_PULSER) );

	if ( sis3807_pulser == (MX_SIS3807_PULSER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannott allocate memory for a MX_SIS3807_PULSER structure"
		"for record '%s'.", record->name );
	}

	record->record_class_struct = pulse_generator;
	record->record_type_struct = sis3807_pulser;

	record->class_specific_function_list =
				&mxd_sis3807_pulse_generator_function_list;

	pulse_generator->record = record;
	sis3807_pulser->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3807_open( MX_RECORD *record )
{
	const char fname[] = "mxd_sis3807_open()";

	MX_PULSE_GENERATOR *pulse_generator;
	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	pulse_generator = (MX_PULSE_GENERATOR *) record->record_class_struct;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
						&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	/* Are we configured for a legal channel number? */

	if ( ( sis3807_pulser->channel_number < 1 )
	  || ( sis3807_pulser->channel_number > sis3807->num_channels ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The channel number %d for 'sis3807_pulser' '%s' is illegal.  "
	"The allowed channel numbers for 'sis3807' record '%s' is 1 to %d.",
			sis3807_pulser->channel_number, record->name,
			sis3807->record->name, sis3807->num_channels );
	}

	/* Set the pulse train finish time to the current time. */

	sis3807_pulser->finish_time = mx_current_clock_tick();

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_sis3807_busy( MX_PULSE_GENERATOR *pulse_generator )
{
	const char fname[] = "mxd_sis3807_busy()";

	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	MX_CLOCK_TICK current_time;
	mx_status_type mx_status;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
							&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.",
			fname, pulse_generator->record->name));

	/* If busy is set to FALSE, then leave it that way. */

	if ( pulse_generator->busy == FALSE )
		return MX_SUCCESSFUL_RESULT;

	/* If we are configured to count forever, then leave the busy flag
	 * alone and return.
	 */

	if ( pulse_generator->num_pulses == MXF_PGN_FOREVER )
		return MX_SUCCESSFUL_RESULT;

	/* See if we have reached the expected finish time of the pulse train.*/

	current_time = mx_current_clock_tick();

	if ( mx_compare_clock_ticks( current_time,
				sis3807_pulser->finish_time ) >= 0 )
	{
		pulse_generator->busy = FALSE;
	} else {
		pulse_generator->busy = TRUE;
	}

	if ( pulse_generator->busy == FALSE ) {
		MX_DEBUG( 2,
	("%s: busy = %d, current_time = (%lu,%lu), finish_time = (%lu,%lu)",
		fname, pulse_generator->busy, current_time.high_order,
		(unsigned long) current_time.low_order,
		sis3807_pulser->finish_time.high_order,
		(unsigned long) sis3807_pulser->finish_time.low_order));

		/* Stop the pulse train. */

		mx_status = mxd_sis3807_stop( pulse_generator );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3807_start( MX_PULSE_GENERATOR *pulse_generator )
{
	const char fname[] = "mxd_sis3807_start()";

	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	MX_CLOCK_TICK start_time, finish_time, countdown_ticks;
	MX_CLOCK_TICK pulse_duration_ticks;
	double total_countdown_time, pulse_time;
	mx_uint32_type enable_channel_value, pulse_register;
	mx_status_type mx_status;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
						&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.",
			fname, pulse_generator->record->name));

	/* Since the SIS3807 has special hardware for generating a
	 * single pulse, we must treat pulse_generator->num_pulses == 1
	 * as a special case.
	 */

	if ( pulse_generator->num_pulses == 1 ) {
		pulse_register = sis3807->base_address;

		switch( sis3807_pulser->channel_number ) {
		case 1:
			pulse_register += MX_SIS3807_PULSE_CHANNEL_1;
			break;
		case 2:
			pulse_register += MX_SIS3807_PULSE_CHANNEL_2;
			break;
		case 3:
			pulse_register += MX_SIS3807_PULSE_CHANNEL_3;
			break;
		case 4:
			pulse_register += MX_SIS3807_PULSE_CHANNEL_4;
			break;
		}

		/* Send the pulse. */

		mx_status = mx_vme_out32( sis3807->vme_record,
						sis3807->crate_number,
						sis3807->address_mode,
						pulse_register,
						1 );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		pulse_generator->busy = TRUE;

		/* Compute when the pulse will be over. */

		start_time = mx_current_clock_tick();

		if ( pulse_generator->mode != MXF_PGN_PULSE ) {
			sis3807_pulser->finish_time = start_time;
		} else {
			pulse_time = 80.0e-9;

			pulse_duration_ticks =
				mx_convert_seconds_to_clock_ticks( pulse_time );

			sis3807_pulser->finish_time = mx_add_clock_ticks(
					start_time, pulse_duration_ticks );
		}

		MX_DEBUG( 2,("%s: single pulse sent.", fname));

		return MX_SUCCESSFUL_RESULT;
	}

	/***** If we get here, then pulse_generator->num_pulses != 1. *****/

	/* Enable the channel. */

	enable_channel_value = 1 << ( 3 + sis3807_pulser->channel_number );

	mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				enable_channel_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	pulse_generator->busy = TRUE;

	/* Return if the pulse generator is configured to count forever. */

	if ( pulse_generator->num_pulses == MXF_PGN_FOREVER )
		return MX_SUCCESSFUL_RESULT;

	/* Record the time when the measurement should be finished. */

	total_countdown_time = pulse_generator->pulse_period
				* (double) pulse_generator->num_pulses;

	/* Give a little extra delay time to try to avoid turning off the
	 * pulse generator early.
	 */

	total_countdown_time += 0.02;		/* Delay time in seconds */

	countdown_ticks
		= mx_convert_seconds_to_clock_ticks( total_countdown_time );

	start_time = mx_current_clock_tick();

	finish_time = mx_add_clock_ticks( start_time, countdown_ticks );

	MX_DEBUG( 2,
		("%s: total_countdown_time = %g, countdown_ticks = (%lu,%lu)",
		fname, total_countdown_time, countdown_ticks.high_order,
			(unsigned long) countdown_ticks.low_order ));

	MX_DEBUG( 2,("%s: start_time = (%lu,%lu), finish_time = (%lu,%lu)",
		fname, start_time.high_order,
		(unsigned long) start_time.low_order,
		finish_time.high_order,
		(unsigned long) finish_time.low_order));

	sis3807_pulser->finish_time = finish_time;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3807_stop( MX_PULSE_GENERATOR *pulse_generator )
{
	const char fname[] = "mxd_sis3807_stop()";

	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	mx_uint32_type disable_channel_value;
	mx_status_type mx_status;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
						&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.",
			fname, pulse_generator->record->name));

	/* Disable the channel. */

	disable_channel_value = 1 << ( 11 + sis3807_pulser->channel_number );

	mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				disable_channel_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	pulse_generator->busy = FALSE;

	/* Set the pulse train finish time to the current time. */

	sis3807_pulser->finish_time = mx_current_clock_tick();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3807_get_parameter( MX_PULSE_GENERATOR *pulse_generator )
{
	const char fname[] = "mxd_sis3807_get_parameter()";

	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	mx_uint32_type status_register_value, mask;
	mx_status_type mx_status;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
						&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s invoked for PULSE_GENERATOR '%s', parameter type '%s' (%d)",
		fname, pulse_generator->record->name,
		mx_get_field_label_string( pulse_generator->record,
					pulse_generator->parameter_type ),
		pulse_generator->parameter_type));

	switch( pulse_generator->parameter_type ) {
	case MXLV_PGN_NUM_PULSES:

		/* The SIS3807 does not have a countdown register, so just
		 * report back the value that we last set.
		 */

		break;
	case MXLV_PGN_PULSE_WIDTH:
		/* The SIS3807 does not provide a way to change the width
		 * of a pulse so compute the value based on the current
		 * number of pulses requested.
		 */

		if ( pulse_generator->mode == MXF_PGN_PULSE ) {
			if ( pulse_generator->num_pulses == 1 ) {
				pulse_generator->pulse_width = 80.0e-9;
			} else {
				pulse_generator->pulse_width = 60.0e-9;
			}
		} else {
			if ( pulse_generator->num_pulses == 1 ) {
				pulse_generator->pulse_width = DBL_MAX;
			} else {
				pulse_generator->pulse_width =
					0.5 * pulse_generator->pulse_period;
			}
		}
		break;
	case MXLV_PGN_PULSE_DELAY:

		/* Pulse delays are not supported for the SIS3807, so
		 * always return 0.
		 */

		pulse_generator->pulse_delay = 0.0;
		break;
	case MXLV_PGN_MODE:
		/* Read the SIS3807 status register. */

		mx_status = mx_vme_in32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_STATUS_REG,
				&status_register_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mask = 1 << ( 15 + sis3807_pulser->channel_number );

		if ( status_register_value & mask ) {
			pulse_generator->mode = MXF_PGN_SQUARE_WAVE;
		} else {
			pulse_generator->mode = MXF_PGN_PULSE;
		}
		break;
	case MXLV_PGN_PULSE_PERIOD:
		/* The pulse preset registers are write only, so just report
		 * the value that we last set.
		 */

		break;
	default:
		return mx_pulse_generator_default_get_parameter_handler(
							pulse_generator );
		break;
	}
	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_sis3807_set_parameter( MX_PULSE_GENERATOR *pulse_generator )
{
	const char fname[] = "mxd_sis3807_set_parameter()";

	MX_SIS3807_PULSER *sis3807_pulser;
	MX_SIS3807 *sis3807;
	double time_quantum, real_preset_value_plus_one;
	mx_uint32_type preset_value, preset_value_plus_one;
	mx_uint32_type control_register_value, preset_register;
	mx_status_type mx_status;

	mx_status = mxd_sis3807_get_pointers( pulse_generator, &sis3807_pulser,
						&sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	control_register_value = 0;

	MX_DEBUG( 2,
	("%s invoked for PULSE_GENERATOR '%s', parameter type '%s' (%d)",
		fname, pulse_generator->record->name,
		mx_get_field_label_string( pulse_generator->record,
					pulse_generator->parameter_type ),
		pulse_generator->parameter_type));

	switch( pulse_generator->parameter_type ) {
	case MXLV_PGN_NUM_PULSES:
		/* Just store the number of pulses locally since the SIS3807
		 * itself does not have a countdown register.
		 */

		break;
	case MXLV_PGN_PULSE_WIDTH:
		/* The SIS3807 does not provide a way to change the width
		 * of a pulse so silently ignore this request and set
		 * the pulse width to a plausible value.
		 */

		if ( pulse_generator->mode == MXF_PGN_PULSE ) {
			if ( pulse_generator->num_pulses == 1 ) {
				pulse_generator->pulse_width = 80.0e-9;
			} else {
				pulse_generator->pulse_width = 60.0e-9;
			}
		} else {
			if ( pulse_generator->num_pulses == 1 ) {
				pulse_generator->pulse_width = DBL_MAX;
			} else {
				pulse_generator->pulse_width =
					0.5 * pulse_generator->pulse_period;
			}
		}
		break;
	case MXLV_PGN_PULSE_DELAY:

		/* Pulse delays are not supported for the SIS3807, so
		 * silently ignore the request.
		 */

		pulse_generator->pulse_delay = 0.0;
		break;
	case MXLV_PGN_MODE:
		switch( pulse_generator->mode ) {
		case MXF_PGN_PULSE:
			control_register_value =
				1 << ( 23 + sis3807_pulser->channel_number );
			break;
		case MXF_PGN_SQUARE_WAVE:
			control_register_value =
				1 << ( 15 + sis3807_pulser->channel_number );
			break;
		default:
			return mx_error( MXE_UNSUPPORTED, fname,
		"Pulse mode %d is not supported for pulse generator '%s'.  "
		"Only pulse mode (1) and square wave mode (2) are supported.",
				pulse_generator->mode,
				pulse_generator->record->name );
			break;
		}
		/* Reprogram the waveform mode. */

		mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				control_register_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXLV_PGN_PULSE_PERIOD:
		if ( sis3807->enable_factor_8_predivider ) {
			time_quantum = 800.0e-9;
		} else {
			time_quantum = 100.0e-9;
		}
		if ( pulse_generator->mode == MXF_PGN_SQUARE_WAVE ) {
			time_quantum *= 2.0;
		}

		real_preset_value_plus_one = mx_divide_safely(
				pulse_generator->pulse_period, time_quantum );

		preset_value_plus_one = mx_round( real_preset_value_plus_one );

		MX_DEBUG( 2,
		("%s: pulse_generator->pulse_period = %g, time_quantum = %g",
			fname, pulse_generator->pulse_period, time_quantum));

		MX_DEBUG( 2,("%s: real_preset_value_plus_one = %g",
			fname, real_preset_value_plus_one));

		MX_DEBUG( 2,("%s: preset_value_plus_one = %lu",
			fname, preset_value_plus_one));

		if ( preset_value_plus_one == 0 ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The pulse period (%g seconds) requested for "
			"pulse generator '%s' is shorter than the "
			"minimum period of %g seconds.",
				pulse_generator->pulse_period,
				pulse_generator->record->name,
				time_quantum );
		}
		if ( preset_value_plus_one >= MX_SIS3807_MAXIMUM_PRESET ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The pulse period (%g seconds) requested for "
			"pulse generator '%s' is longer than the "
			"maximum period of %g seconds.",
				pulse_generator->pulse_period,
				pulse_generator->record->name,
				time_quantum * (double)
					( 1 + MX_SIS3807_MAXIMUM_PRESET ) );
		}

		preset_value = preset_value_plus_one - 1;

		/* Figure out which register to write the preset_value to. */

		preset_register = sis3807->base_address;

		switch( sis3807_pulser->channel_number ) {
		case 1:
			preset_register += MX_SIS3807_PRESET_FACTOR_PULSER_1;
			break;
		case 2:
			preset_register += MX_SIS3807_PRESET_FACTOR_PULSER_2;
			break;
		case 3:
			preset_register += MX_SIS3807_PRESET_FACTOR_PULSER_3;
			break;
		case 4:
			preset_register += MX_SIS3807_PRESET_FACTOR_PULSER_4;
			break;
		}

		/* Update the preset. */

		mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				preset_register,
				preset_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	default:
		return mx_pulse_generator_default_set_parameter_handler(
							pulse_generator );
		break;
	}
	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

