/*
 * Name:    i_sis3807.c
 *
 * Purpose: MX MCS driver for the Struck SIS3807 multichannel pulse generator.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_vme.h"

#include "i_sis3807.h"

MX_RECORD_FUNCTION_LIST mxi_sis3807_record_function_list = {
	NULL,
	mxi_sis3807_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxi_sis3807_open,
	NULL,
	NULL,
	mxi_sis3807_open
};

MX_RECORD_FIELD_DEFAULTS mxi_sis3807_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_SIS3807_STANDARD_FIELDS
};

long mxi_sis3807_num_record_fields
		= sizeof( mxi_sis3807_record_field_defaults )
			/ sizeof( mxi_sis3807_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_sis3807_rfield_def_ptr
			= &mxi_sis3807_record_field_defaults[0];

/* --- */

static mx_status_type
mxi_sis3807_get_pointers( MX_RECORD *sis3807_record,
			MX_SIS3807 **sis3807,
			const char *calling_fname )
{
	static const char fname[] = "mxi_sis3807_get_pointers()";

	if ( sis3807_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_RECORD pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( sis3807 == (MX_SIS3807 **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_SIS3807 pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*sis3807 = (MX_SIS3807 *) sis3807_record->record_type_struct;

	if ( *sis3807 == (MX_SIS3807 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SIS3807 pointer for record '%s' is NULL.",
			sis3807_record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxi_sis3807_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxi_sis3807_create_record_structures()";

	MX_SIS3807 *sis3807;

	sis3807 = (MX_SIS3807 *) malloc( sizeof(MX_SIS3807) );

	if ( sis3807 == (MX_SIS3807 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_SIS3807 structure." );
	}

	record->record_class_struct = NULL;
	record->class_specific_function_list = NULL;

	record->record_type_struct = sis3807;

	sis3807->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_sis3807_open( MX_RECORD *record )
{
	static const char fname[] = "mxi_sis3807_open()";

	MX_SIS3807 *sis3807;
	int i;
	mx_uint32_type module_id_register, control_register, preset_register;
	mx_status_type mx_status;

	mx_status = mxi_sis3807_get_pointers( record, &sis3807, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Parse the address mode string. */

	MX_DEBUG( 2,("%s: sis3807->address_mode_name = '%s'",
			fname, sis3807->address_mode_name));

	mx_status = mx_vme_parse_address_mode( sis3807->address_mode_name,
						&(sis3807->address_mode) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: sis3807->address_mode = %lu",
			fname, sis3807->address_mode));

	/* Reset the SIS3807. */

	mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_RESET_REG,
				1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Read the module identification register. */

	mx_status = mx_vme_in32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
			sis3807->base_address + MX_SIS3807_MODULE_ID_REG,
				&module_id_register );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sis3807->module_id = ( module_id_register >> 16 ) & 0xffff;
	sis3807->firmware_version = ( module_id_register >> 12 ) & 0xf;

	/* Apparently, the documentation concerning SIS3807 firmware version
	 * numbers at the end of section 7.3 is incorrect.  You must add 1
	 * to the value obtained from the module id register to get the
	 * firmware version number.
	 */

	(sis3807->firmware_version)++;

	MX_DEBUG( 2,("%s: module id = %#lx, firmware version = %lu",
		fname, sis3807->module_id, sis3807->firmware_version));

	/* After a reset, most of the control register bits are set the
	 * way that we want.  The only things we want to change are:
	 * 1.  Select 2 channel mode if requested.
	 * 2.  Turn on the factor 8 predivider if requested.
	 * 3.  Turn on the general enable bit.
	 *
	 * The individual channels will not start pulsing until their
	 * individual enable bits are turned on.
	 */

	control_register = 0;

	/* See whether we are configured for 2 channels or 4 channels. */

	if ( sis3807->num_channels == 4 ) {

		/* The SIS3807 will default to four channels after a reset
		 * so we do not need to do anything here.
		 */

	} else if ( sis3807->num_channels == 2 ) {

		/* Firmware version 1 does not support 2 channel mode.*/

		if ( sis3807->firmware_version == 1 ) {
			return mx_error( MXE_UNSUPPORTED, fname,
			"SIS3807 firmware version 1 does not support 2 channel "
			"mode.  Two channel mode requires firmware version 2 "
			"or later." );

		} else {
			/* Turn on the 'set channel 2 mode' bit. */

			control_register |= MXF_SIS3807_SET_2_CHANNEL_MODE_BIT;
		}
	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
	"The SIS3807 module '%s' only supports 2 channel and 4 channel mode.  "
	"You have attempted to select %d channels which is not supported.",
			record->name, sis3807->num_channels );
	}

	/* Turn on the factor 8 predivider if requested. */

	if ( sis3807->enable_factor_8_predivider ) {
		control_register |=
			MXF_SIS3807_ENABLE_COMMON_FACTOR_8_PREDIVIDER_BIT;
	}

	/* Program the control register. */

	mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				control_register );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set all the pulser presets to 0.  This means that the default
	 * pulse period is as small as possible.
	 */

	for ( i = 0; i < sis3807->num_channels; i++ ) {

		preset_register = sis3807->base_address + 0x40 + 0x4 * i;

		mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				preset_register,
				0 );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Turn on the general enable bit. */

	mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				MXF_SIS3807_GENERAL_ENABLE_BIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Initialize the output invert register. */

	if ( sis3807->firmware_version == 1 ) {
		/* Firmware version 1 does not have an output invert register.
		 * Instead we invert outputs by setting bit 2 in the control
		 * register.  In later versions of the firmware, bit 2 is
		 * reused as the "set 2 channel mode bit" which is what we
		 * call it here.
		 */

		if ( sis3807->output_invert_register ) {
			mx_status = mx_vme_out32( sis3807->vme_record,
				sis3807->crate_number,
				sis3807->address_mode,
				sis3807->base_address + MX_SIS3807_CONTROL_REG,
				MXF_SIS3807_SET_2_CHANNEL_MODE_BIT );
		}
	} else {
		/* Firmware versions 2 and after have
		 * the output invert register.
		 */

		mx_status = mx_vme_out32( sis3807->vme_record,
					sis3807->crate_number,
					sis3807->address_mode,
			sis3807->base_address + MX_SIS3807_OUTPUT_INVERT_REG,
					sis3807->output_invert_register );
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

